/*
 * Copyright (c) 2001-2004 The Trustees of Indiana University.  
 *                         All rights reserved.
 * Copyright (c) 1998-2001 University of Notre Dame. 
 *                         All rights reserved.
 * Copyright (c) 1994-1998 The Ohio State University.  
 *                         All rights reserved.
 * 
 * This file is part of the LAM/MPI software package.  For license
 * information, see the LICENSE file in the top level directory of the
 * LAM/MPI source distribution.
 * 
 * $HEADER$
 *
 * $Id: alltoallv_inter.c,v 1.1 2004/03/15 01:18:16 adharurk Exp $
 *
 * Program to test Intercommunicator Alltoallv
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <mpi.h> 
#include <lamtest_error.h>

#define MAXLEN 10000

static void intercomm_alltoallv(MPI_Comm inter, int rank);

int main(int argc, char **argv ) 
{ 
  int size, rank, port_length, color, local_rank;
  MPI_Comm comm, intercomm;
  char *port_name;
  int buf = 1;

  MPI_Init(&argc, &argv);

  lamtest_need_even(__FILE__, __LINE__);
  lamtest_check_size(__FILE__, __LINE__, 2, 1);

  MPI_Comm_size(MPI_COMM_WORLD, &size);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  /* Even rank gets 1 while odd ranks get 0 value for color */
  color = (rank % 2) ? 0 : 1;
  MPI_Comm_split(MPI_COMM_WORLD, color, rank, &comm);
  
  if (color) {
    /* even ranks */
    MPI_Comm_rank (comm, &local_rank);
    if (local_rank == 0) {
      port_name = malloc(MPI_MAX_PORT_NAME * sizeof(char));
      MPI_Open_port(MPI_INFO_NULL, port_name);
      port_length = strlen(port_name) + 1;

      MPI_Send(&port_length, 1, MPI_INT, rank + 1, 1, MPI_COMM_WORLD);
      MPI_Send(port_name, port_length, MPI_CHAR, rank + 1, 1, MPI_COMM_WORLD);
    }      

    MPI_Comm_accept(port_name, MPI_INFO_NULL, 0, comm, &intercomm);
    intercomm_alltoallv(intercomm, local_rank);
    MPI_Comm_disconnect(&intercomm);

    if (local_rank == 0) {
      MPI_Close_port(port_name);
      free(port_name);
    }

  } else {
    /* odd ranks */
    MPI_Comm_rank(comm, &local_rank);
    if (local_rank == 0) {
      MPI_Recv(&port_length, 1, MPI_INT, rank - 1, 1, MPI_COMM_WORLD, 
	       MPI_STATUS_IGNORE);
      
      /* allocate memory for the the port name */
      port_name = malloc(port_length * sizeof(char));
      MPI_Recv(port_name, port_length, MPI_CHAR, rank - 1, 1, MPI_COMM_WORLD, 
	       MPI_STATUS_IGNORE);
    }      

    /* connect to the server */
    MPI_Comm_connect(port_name, MPI_INFO_NULL, 0, comm, &intercomm);
    intercomm_alltoallv(intercomm, local_rank);
    MPI_Comm_disconnect(&intercomm);

    if (local_rank == 0)
      free(port_name);
  }
  MPI_Comm_free(&comm);
  MPI_Finalize();
  return 0;
} 

static void
intercomm_alltoallv(MPI_Comm inter, int rank)
{
  MPI_Comm comm;
  int *sbuf, *rbuf;
  int size;
  int *sendcounts, *recvcounts, *rdispls, *sdispls;
  int i, j, *p;
  MPI_Datatype *sdtypes, *rdtypes;
  
  /* Create the buffer */
  MPI_Comm_size(inter, &size );
  sbuf = (int *)malloc( size * size * sizeof(int) );
  rbuf = (int *)malloc( size * size * sizeof(int) );
  if (!sbuf || !rbuf) {
    lamtest_error(__FILE__, __LINE__, "Doh!  Rank %d was not able to allocateeough memory.  MPI test aborted!\n", rank);
  }
  
  /* Initialize buffers */
  for (i=0; i<size*size; i++) {
    sbuf[i] = i + 100*rank;
    rbuf[i] = -i;
  }
  
  /* Create and load the arguments to alltoallw */
  sendcounts = (int *)malloc( size * sizeof(int) );
  recvcounts = (int *)malloc( size * sizeof(int) );
  rdispls    = (int *)malloc( size * sizeof(int) );
  sdispls    = (int *)malloc( size * sizeof(int) );
  sdtypes    = (MPI_Datatype *) malloc(size * sizeof(MPI_Datatype));
  rdtypes    = (MPI_Datatype *) malloc(size * sizeof(MPI_Datatype));
  if (!sendcounts || !recvcounts || !rdispls || !sdispls) {
    lamtest_error(__FILE__, __LINE__, "Doh!  Rank %d was not able to allocateenough memory.  MPI test aborted!\n", rank);
  }
  for (i=0; i<size; i++) {
    sendcounts[i] = i;
    recvcounts[i] = rank;
    rdispls[i]    = i * rank;
    sdispls[i]    = ((i * (i+1))/2);
  }
  MPI_Alltoallv( sbuf, sendcounts, sdispls, MPI_INT,
		 rbuf, recvcounts, rdispls, MPI_INT, inter );
     
  /* Check rbuf */
  for (i=0; i<size; i++) {
    p = rbuf + rdispls[i];
    for (j=0; j<rank; j++) {
      if (p[j] != i * 100 + (rank*(rank+1))/2 + j) {
	lamtest_error(__FILE__, __LINE__, "bad answer (%d) (should be %d)\n",p[j], i * 100 + (rank*(rank+1))/2 + j);
      }
    }
  }

  free( sdispls );
  free( rdispls );
  free( recvcounts );
  free( sendcounts );
  free( rbuf );
  free( sbuf );
}

